@@ -71,9 +71,32 @@ static krb5_enctype v51_earlier_enctypes[] = {
 	0
 };
 
+/* The following list containst all attributes handled by adcli, some are
+ * read-only and the others can be written as well. To properly document the
+ * required permissions each attribute which adcli tries to modify should have
+ * a comment starting with ':ADPermissions:' and the related permissions in AD
+ * on the same line. Multiple permissions can be seperated with a '*'. For all
+ * other attribute a suitable comment is very welcome. */
+static char *default_ad_ldap_attrs[] =  {
+	"sAMAccountName", /* Only set during creation */
+	"userPrincipalName",   /* :ADPermissions: Read/Write userPrincipal Name */
+	"msDS-KeyVersionNumber", /* Manages by AD */
+	"msDS-supportedEncryptionTypes", /* :ADPermissions: Read/Write msDS-SupportedEncryptionTypes */
+	"dNSHostName", /* :ADPermissions: Read/Write dNSHostName * Read and write DNS host name attributes * Validated write to DNS host name */
+	"servicePrincipalName", /* :ADPermissions: Read/Write servicePrincipalName * Validated write to service principal name */
+	"operatingSystem", /* :ADPermissions: Read/Write Operating System */
+	"operatingSystemVersion", /* :ADPermissions: Read/Write Operating System Version */
+	"operatingSystemServicePack", /* :ADPermissions: Read/Write operatingSystemServicePack */
+	"pwdLastSet", /* Managed by AD */
+	"userAccountControl", /* :ADPermissions: Read/Write userAccountControl */
+	"description", /* :ADPermissions: Read/Write Description */
+	NULL,
+};
+
 /* Some constants for the userAccountControl AD LDAP attribute, see e.g.
  * https://support.microsoft.com/en-us/help/305144/how-to-use-the-useraccountcontrol-flags-to-manipulate-user-account-pro
  * for details. */
+#define UAC_ACCOUNTDISABLE             0x0002
 #define UAC_WORKSTATION_TRUST_ACCOUNT  0x1000
 #define UAC_DONT_EXPIRE_PASSWORD      0x10000
 #define UAC_TRUSTED_FOR_DELEGATION    0x80000
@@ -81,6 +104,8 @@ static krb5_enctype v51_earlier_enctypes[] = {
 struct _adcli_enroll {
 	int refs;
 	adcli_conn *conn;
+	bool is_service;
+	bool is_service_explicit;
 
 	char *host_fqdn;
 	int host_fqdn_explicit;
@@ -128,8 +153,31 @@ struct _adcli_enroll {
 	char *samba_data_tool;
 	bool trusted_for_delegation;
 	int trusted_for_delegation_explicit;
+	bool account_disable;
+	int account_disable_explicit;
+	char *description;
 };
 
+static const char *
+s_or_c (adcli_enroll *enroll)
+{
+	return enroll->is_service ? "service" : "computer";
+}
+
+static void
+check_if_service (adcli_enroll *enroll,
+                  LDAP *ldap,
+                  LDAPMessage *results)
+{
+	char **objectclasses = NULL;
+
+	objectclasses = _adcli_ldap_parse_values (ldap, results, "objectClass");
+	enroll->is_service = _adcli_strv_has_ex (objectclasses,
+	                                         "msDS-ManagedServiceAccount",
+	                                         strcasecmp) == 1 ? true : false;
+	_adcli_strv_free (objectclasses);
+}
+
 static adcli_result
 ensure_host_fqdn (adcli_result res,
                   adcli_enroll *enroll)
@@ -164,13 +212,15 @@ ensure_computer_name (adcli_result res,
 		return res;
 
 	if (enroll->computer_name) {
-		_adcli_info ("Enrolling computer name: %s",
+		_adcli_info ("Enrolling %s name: %s",
+		             s_or_c (enroll),
 		             enroll->computer_name);
 		return ADCLI_SUCCESS;
 	}
 
 	if (!enroll->host_fqdn) {
-		_adcli_err ("No host name from which to determine the computer name");
+		_adcli_err ("No host name from which to determine the %s name",
+		            s_or_c (enroll));
 		return ADCLI_ERR_CONFIG;
 	}
 
@@ -212,6 +262,29 @@ ensure_computer_sam (adcli_result res,
 	return ADCLI_SUCCESS;
 }
 
+typedef int (rand_filter) (char *password, int length);
+
+static int
+filter_sam_chars (char *password,
+                       int length)
+{
+	int i, j;
+
+	/*
+	 * There are a couple of restrictions for characters in the
+	 * sAMAccountName attribute value, for our purpose (random suffix)
+	 * letters and numbers are sufficient.
+	 */
+	for (i = 0, j = 0; i < length; i++) {
+		if (password[i] >= 48 && password[i] <= 122 &&
+		    isalnum (password[i]))
+			password[j++] = password[i];
+	}
+
+	/* return the number of valid characters remaining */
+	return j;
+}
+
 static int
 filter_password_chars (char *password,
                        int length)
@@ -236,7 +309,8 @@ filter_password_chars (char *password,
 
 static char *
 generate_host_password  (adcli_enroll *enroll,
-                         size_t length)
+                         size_t length,
+                         rand_filter *filter)
 {
 	char *password;
 	krb5_context k5;
@@ -258,7 +332,7 @@ generate_host_password  (adcli_enroll *enroll,
 		code = krb5_c_random_make_octets (k5, &buffer);
 		return_val_if_fail (code == 0, NULL);
 
-		at += filter_password_chars (buffer.data, buffer.length);
+		at += filter (buffer.data, buffer.length);
 		assert (at <= length);
 	}
 
@@ -286,7 +360,7 @@ ensure_computer_password (adcli_result res,
 		_adcli_info ("Using default reset computer password");
 
 	} else {
-		enroll->computer_password = generate_host_password (enroll, length);
+		enroll->computer_password = generate_host_password (enroll, length, filter_password_chars);
 		return_unexpected_if_fail (enroll->computer_password != NULL);
 		_adcli_info ("Generated %d character computer password", length);
 	}
@@ -446,13 +520,15 @@ ensure_keytab_principals (adcli_result res,
 {
 	krb5_context k5;
 	krb5_error_code code;
-	int count;
+	int count = 0;
 	int at, i;
 
 	/* Prepare the principals we're going to add to the keytab */
 
-	return_unexpected_if_fail (enroll->service_principals);
-	count = _adcli_strv_len (enroll->service_principals);
+	if (!enroll->is_service) {
+		return_unexpected_if_fail (enroll->service_principals);
+		count = _adcli_strv_len (enroll->service_principals);
+	}
 
 	k5 = adcli_conn_get_krb5_context (enroll->conn);
 	return_unexpected_if_fail (k5 != NULL);
@@ -531,8 +607,12 @@ static adcli_result
 lookup_computer_container (adcli_enroll *enroll,
                            LDAP *ldap)
 {
-	char *attrs[] = { "wellKnownObjects", NULL };
-	char *prefix = "B:32:AA312825768811D1ADED00C04FD8D5CD:";
+	char *attrs[] = { enroll->is_service ? "otherWellKnownObjects"
+	                                     : "wellKnownObjects", NULL };
+	const char *prefix = enroll->is_service ? "B:32:1EB93889E40C45DF9F0C64D23BBB6237:"
+	                                        : "B:32:AA312825768811D1ADED00C04FD8D5CD:";
+	const char *filter = enroll->is_service ? "(&(objectClass=container)(cn=Managed Service Accounts))"
+	                                        : "(&(objectClass=container)(cn=Computers))";
 	int prefix_len;
 	LDAPMessage *results;
 	const char *base;
@@ -558,10 +638,11 @@ lookup_computer_container (adcli_enroll *enroll,
 
 	} else if (ret != LDAP_SUCCESS) {
 		return _adcli_ldap_handle_failure (ldap, ADCLI_ERR_DIRECTORY,
-		                                   "Couldn't lookup computer container: %s", base);
+		                                   "Couldn't lookup %s container: %s",
+		                                   s_or_c (enroll), base);
 	}
 
-	values = _adcli_ldap_parse_values (ldap, results, "wellKnownObjects");
+	values = _adcli_ldap_parse_values (ldap, results, attrs[0]);
 	ldap_msgfree (results);
 
 	prefix_len = strlen (prefix);
@@ -569,8 +650,8 @@ lookup_computer_container (adcli_enroll *enroll,
 		if (strncmp (values[i], prefix, prefix_len) == 0) {
 			enroll->computer_container = strdup (values[i] + prefix_len);
 			return_unexpected_if_fail (enroll->computer_container != NULL);
-			_adcli_info ("Found well known computer container at: %s",
-			             enroll->computer_container);
+			_adcli_info ("Found well known %s container at: %s",
+			             s_or_c (enroll), enroll->computer_container);
 			break;
 		}
 	}
@@ -579,14 +660,14 @@ lookup_computer_container (adcli_enroll *enroll,
 
 	/* Try harder */
 	if (!enroll->computer_container) {
-		ret = ldap_search_ext_s (ldap, base, LDAP_SCOPE_BASE,
-		                         "(&(objectClass=container)(cn=Computers))",
+		ret = ldap_search_ext_s (ldap, base, LDAP_SCOPE_BASE, filter,
 		                         attrs, 0, NULL, NULL, NULL, -1, &results);
 		if (ret == LDAP_SUCCESS) {
 			enroll->computer_container = _adcli_ldap_parse_dn (ldap, results);
 			if (enroll->computer_container) {
-				_adcli_info ("Well known computer container not "
+				_adcli_info ("Well known %s container not "
 				             "found, but found suitable one at: %s",
+				             s_or_c (enroll),
 				             enroll->computer_container);
 			}
 		}
@@ -602,7 +683,8 @@ lookup_computer_container (adcli_enroll *enroll,
 	}
 
 	if (!enroll->computer_container) {
-		_adcli_err ("Couldn't find location to create computer accounts");
+		_adcli_err ("Couldn't find location to create %s accounts",
+		            s_or_c (enroll));
 		return ADCLI_ERR_DIRECTORY;
 	}
 
@@ -630,7 +712,8 @@ calculate_computer_account (adcli_enroll *enroll,
 	if (asprintf (&enroll->computer_dn, "CN=%s,%s", enroll->computer_name, enroll->computer_container) < 0)
 		return_unexpected_if_reached ();
 
-	_adcli_info ("Calculated computer account: %s", enroll->computer_dn);
+	_adcli_info ("Calculated %s account: %s",
+	             s_or_c (enroll), enroll->computer_dn);
 	return ADCLI_SUCCESS;
 }
 
@@ -722,7 +805,7 @@ static adcli_result
 create_computer_account (adcli_enroll *enroll,
                          LDAP *ldap)
 {
-	char *vals_objectClass[] = { "computer", NULL };
+	char *vals_objectClass[] = { enroll->is_service ? "msDS-ManagedServiceAccount" : "computer", NULL };
 	LDAPMod objectClass = { LDAP_MOD_ADD, "objectClass", { vals_objectClass, } };
 	char *vals_sAMAccountName[] = { enroll->computer_sam, NULL };
 	LDAPMod sAMAccountName = { LDAP_MOD_ADD, "sAMAccountName", { vals_sAMAccountName, } };
@@ -741,6 +824,8 @@ create_computer_account (adcli_enroll *enroll,
 	char *vals_userPrincipalName[] = { enroll->user_principal, NULL };
 	LDAPMod userPrincipalName = { LDAP_MOD_ADD, "userPrincipalName", { vals_userPrincipalName, }, };
 	LDAPMod servicePrincipalName = { LDAP_MOD_ADD, "servicePrincipalName", { enroll->service_principals, } };
+	char *vals_description[] = { enroll->description, NULL };
+	LDAPMod description = { LDAP_MOD_ADD, "description", { vals_description, }, };
 
 	char *val = NULL;
 
@@ -759,6 +844,7 @@ create_computer_account (adcli_enroll *enroll,
 		&operatingSystemServicePack,
 		&userPrincipalName,
 		&servicePrincipalName,
+		&description,
 		NULL
 	};
 
@@ -778,7 +864,7 @@ create_computer_account (adcli_enroll *enroll,
 	m = 0;
 	for (c = 0; c < mods_count - 1; c++) {
 		/* Skip empty LDAP sttributes */
-		if (all_mods[c]->mod_vals.modv_strvals[0] != NULL) {
+		if (all_mods[c]->mod_vals.modv_strvals != NULL && all_mods[c]->mod_vals.modv_strvals[0] != NULL) {
 			mods[m++] = all_mods[c];
 		}
 	}
@@ -814,7 +900,8 @@ create_computer_account (adcli_enroll *enroll,
 		                                   enroll->computer_dn);
 	}
 
-	_adcli_info ("Created computer account: %s", enroll->computer_dn);
+	_adcli_info ("Created %s account: %s", s_or_c (enroll),
+	                                       enroll->computer_dn);
 	return ADCLI_SUCCESS;
 }
 
@@ -861,17 +948,17 @@ validate_computer_account (adcli_enroll *enroll,
 	assert (enroll->computer_dn != NULL);
 
 	if (already_exists && !allow_overwrite) {
-		_adcli_err ("The computer account %s already exists",
-		            enroll->computer_name);
+		_adcli_err ("The %s account %s already exists",
+		            s_or_c (enroll), enroll->computer_name);
 		return ADCLI_ERR_CONFIG;
 	}
 
 	/* Do we have an explicitly requested ou? */
 	if (enroll->domain_ou && enroll->domain_ou_explicit && already_exists) {
 		if (!_adcli_ldap_dn_has_ancestor (enroll->computer_dn, enroll->domain_ou)) {
-			_adcli_err ("The computer account %s already exists, "
+			_adcli_err ("The %s account %s already exists, "
 			            "but is not in the desired organizational unit.",
-			            enroll->computer_name);
+			            s_or_c (enroll), enroll->computer_name);
 			return ADCLI_ERR_CONFIG;
 		}
 	}
@@ -896,7 +983,8 @@ delete_computer_account (adcli_enroll *enroll,
 		                                   "Couldn't delete computer account: %s",
 		                                   enroll->computer_dn);
 	} else {
-		_adcli_info ("Deleted computer account at: %s", enroll->computer_dn);
+		_adcli_info ("Deleted %s account at: %s", s_or_c (enroll),
+		                                          enroll->computer_dn);
 	}
 
 	return ADCLI_SUCCESS;
@@ -905,10 +993,11 @@ delete_computer_account (adcli_enroll *enroll,
 static adcli_result
 locate_computer_account (adcli_enroll *enroll,
                          LDAP *ldap,
+                         bool use_fqdn,
                          LDAPMessage **rresults,
                          LDAPMessage **rentry)
 {
-	char *attrs[] = { "1.1", NULL };
+	char *attrs[] = { "objectClass", "CN", NULL };
 	LDAPMessage *results = NULL;
 	LDAPMessage *entry = NULL;
 	const char *base;
@@ -918,10 +1007,22 @@ locate_computer_account (adcli_enroll *enroll,
 	int ret = 0;
 
 	/* If we don't yet know our computer dn, then try and find it */
-	value = _adcli_ldap_escape_filter (enroll->computer_sam);
-	return_unexpected_if_fail (value != NULL);
-	if (asprintf (&filter, "(&(objectClass=computer)(sAMAccountName=%s))", value) < 0)
-		return_unexpected_if_reached ();
+	if (use_fqdn) {
+		return_unexpected_if_fail (enroll->host_fqdn != NULL);
+		value = _adcli_ldap_escape_filter (enroll->host_fqdn);
+		return_unexpected_if_fail (value != NULL);
+		if (asprintf (&filter, "(&(objectClass=%s)(dNSHostName=%s))",
+		              enroll->is_service ? "msDS-ManagedServiceAccount" : "computer",
+		              value) < 0)
+			return_unexpected_if_reached ();
+	} else {
+		value = _adcli_ldap_escape_filter (enroll->computer_sam);
+		return_unexpected_if_fail (value != NULL);
+		if (asprintf (&filter, "(&(objectClass=%s)(sAMAccountName=%s))",
+		              enroll->is_service ? "msDS-ManagedServiceAccount" : "computer",
+		              value) < 0)
+			return_unexpected_if_reached ();
+	}
 	free (value);
 
 	base = adcli_conn_get_default_naming_context (enroll->conn);
@@ -934,27 +1035,36 @@ locate_computer_account (adcli_enroll *enroll,
 	if (ret == LDAP_SUCCESS) {
 		entry = ldap_first_entry (ldap, results);
 
-		/* If we found a computer account, make note of dn */
+		/* If we found a computer/service account, make note of dn */
 		if (entry) {
+			if (!enroll->is_service_explicit) {
+				check_if_service ( enroll, ldap, results);
+			}
 			dn = ldap_get_dn (ldap, entry);
 			free (enroll->computer_dn);
 			enroll->computer_dn = strdup (dn);
 			return_unexpected_if_fail (enroll->computer_dn != NULL);
-			_adcli_info ("Found computer account for %s at: %s",
-			             enroll->computer_sam, dn);
+			_adcli_info ("Found %s account for %s at: %s",
+			             s_or_c (enroll),
+			             use_fqdn ? enroll->host_fqdn
+			                      : enroll->computer_sam, dn);
 			ldap_memfree (dn);
 
 		} else {
 			ldap_msgfree (results);
 			results = NULL;
-			_adcli_info ("Computer account for %s does not exist",
-			             enroll->computer_sam);
+			_adcli_info ("A %s account for %s does not exist",
+			             s_or_c (enroll),
+			             use_fqdn ? enroll->host_fqdn
+			                      : enroll->computer_sam);
 		}
 
 	} else {
 		return _adcli_ldap_handle_failure (ldap, ADCLI_ERR_DIRECTORY,
-		                                   "Couldn't lookup computer account: %s",
-		                                   enroll->computer_sam);
+		                                   "Couldn't lookup %s account: %s",
+		                                   s_or_c (enroll),
+		                                   use_fqdn ? enroll->host_fqdn
+		                                            :enroll->computer_sam);
 	}
 
 	if (rresults)
@@ -975,7 +1085,7 @@ load_computer_account (adcli_enroll *enroll,
                        LDAPMessage **rresults,
                        LDAPMessage **rentry)
 {
-	char *attrs[] = { "1.1", NULL };
+	char *attrs[] = { "objectClass", NULL };
 	LDAPMessage *results = NULL;
 	LDAPMessage *entry = NULL;
 	int ret;
@@ -987,7 +1097,9 @@ load_computer_account (adcli_enroll *enroll,
 	if (ret == LDAP_SUCCESS) {
 		entry = ldap_first_entry (ldap, results);
 		if (entry) {
-			_adcli_info ("Found computer account for %s at: %s",
+			check_if_service (enroll, ldap, results);
+			_adcli_info ("Found %s account for %s at: %s",
+			             s_or_c (enroll),
 			             enroll->computer_sam, enroll->computer_dn);
 		}
 
@@ -1012,6 +1124,59 @@ load_computer_account (adcli_enroll *enroll,
 	return ADCLI_SUCCESS;
 }
 
+static adcli_result
+refresh_service_account_name_sam_and_princ (adcli_enroll *enroll,
+                                            const char *name)
+{
+	adcli_result res;
+
+	adcli_enroll_set_computer_name (enroll, name);
+	res = ensure_computer_sam (ADCLI_SUCCESS, enroll);
+	res = ensure_keytab_principals (res, enroll);
+
+	return res;
+}
+
+static adcli_result
+calculate_random_service_account_name (adcli_enroll *enroll)
+{
+	char *suffix;
+	char *new_name;
+	int ret;
+	adcli_result res;
+
+	suffix = generate_host_password (enroll, 3, filter_sam_chars);
+	return_unexpected_if_fail (suffix != NULL);
+
+	ret = asprintf (&new_name, "%s!%s", enroll->computer_name, suffix);
+	free (suffix);
+	return_unexpected_if_fail (ret > 0);
+
+	res = refresh_service_account_name_sam_and_princ (enroll, new_name);
+	free (new_name);
+
+	return res;
+}
+
+static adcli_result
+get_service_account_name_from_ldap (adcli_enroll *enroll, LDAPMessage *results)
+{
+	LDAP *ldap;
+	char *cn;
+	adcli_result res;
+
+	ldap = adcli_conn_get_ldap_connection (enroll->conn);
+	assert (ldap != NULL);
+
+	cn = _adcli_ldap_parse_value (ldap, results, "CN");
+	return_unexpected_if_fail (cn != NULL);
+
+	res = refresh_service_account_name_sam_and_princ (enroll, cn);
+	free (cn);
+
+	return res;
+}
+
 static adcli_result
 locate_or_create_computer_account (adcli_enroll *enroll,
                                    int allow_overwrite)
@@ -1027,14 +1192,39 @@ locate_or_create_computer_account (adcli_enroll *enroll,
 
 	/* Try to find the computer account */
 	if (!enroll->computer_dn) {
-		res = locate_computer_account (enroll, ldap, &results, &entry);
+		res = locate_computer_account (enroll, ldap, false,
+		                               &results, &entry);
 		if (res != ADCLI_SUCCESS)
 			return res;
 		searched = 1;
 	}
 
+	/* Try with fqdn for service accounts */
+	if (!enroll->computer_dn && enroll->is_service
+	                && enroll->host_fqdn != NULL) {
+		res = locate_computer_account (enroll, ldap, true,
+		                               &results, &entry);
+		if (res != ADCLI_SUCCESS)
+			return res;
+		searched = 1;
+
+		if (results != NULL) {
+			res = get_service_account_name_from_ldap (enroll,
+			                                          results);
+			if (res != ADCLI_SUCCESS) {
+				return res;
+			}
+		}
+	}
+
 	/* Next try and come up with where we think it should be */
 	if (enroll->computer_dn == NULL) {
+		if (enroll->is_service && !enroll->computer_name_explicit) {
+			res = calculate_random_service_account_name (enroll);
+			if (res != ADCLI_SUCCESS) {
+				return res;
+			}
+		}
 		res = calculate_computer_account (enroll, ldap);
 		if (res != ADCLI_SUCCESS)
 			return res;
@@ -1053,6 +1243,12 @@ locate_or_create_computer_account (adcli_enroll *enroll,
 	if (res == ADCLI_SUCCESS && entry == NULL)
 		res = create_computer_account (enroll, ldap);
 
+	/* Service account already exists, just continue and update the
+	 * password */
+	if (enroll->is_service && entry != NULL) {
+		res = ADCLI_SUCCESS;
+	}
+
 	if (results)
 		ldap_msgfree (results);
 
@@ -1088,7 +1284,8 @@ set_password_with_user_creds (adcli_enroll *enroll)
 	                                       &result_code_string, &result_string);
 
 	if (code != 0) {
-		_adcli_err ("Couldn't set password for computer account: %s: %s",
+		_adcli_err ("Couldn't set password for %s account: %s: %s",
+		            s_or_c (enroll),
 		            enroll->computer_sam, krb5_get_error_message (k5, code));
 		/* TODO: Parse out these values */
 		res = ADCLI_ERR_DIRECTORY;
@@ -1102,7 +1299,8 @@ set_password_with_user_creds (adcli_enroll *enroll)
 		if (result_string.length)
 			message = _adcli_str_dupn (result_string.data, result_string.length);
 #endif
-		_adcli_err ("Cannot set computer password: %.*s%s%s",
+		_adcli_err ("Cannot set %s password: %.*s%s%s",
+		            s_or_c (enroll),
 		            (int)result_code_string.length, result_code_string.data,
 		            message ? ": " : "", message ? message : "");
 		res = ADCLI_ERR_CREDENTIALS;
@@ -1112,7 +1310,7 @@ set_password_with_user_creds (adcli_enroll *enroll)
 		free (message);
 #endif
 	} else {
-		_adcli_info ("Set computer password");
+		_adcli_info ("Set %s password", s_or_c (enroll));
 		if (enroll->kvno > 0) {
 			enroll->kvno++;
 			_adcli_info ("kvno incremented to %d", enroll->kvno);
@@ -1145,7 +1343,8 @@ set_password_with_computer_creds (adcli_enroll *enroll)
 
 	code = _adcli_kinit_computer_creds (enroll->conn, "kadmin/changepw", NULL, &creds);
 	if (code != 0) {
-		_adcli_err ("Couldn't get change password ticket for computer account: %s: %s",
+		_adcli_err ("Couldn't get change password ticket for %s account: %s: %s",
+		            s_or_c (enroll),
 		            enroll->computer_sam, krb5_get_error_message (k5, code));
 		return ADCLI_ERR_DIRECTORY;
 	}
@@ -1156,7 +1355,8 @@ set_password_with_computer_creds (adcli_enroll *enroll)
 	krb5_free_cred_contents (k5, &creds);
 
 	if (code != 0) {
-		_adcli_err ("Couldn't change password for computer account: %s: %s",
+		_adcli_err ("Couldn't change password for %s account: %s: %s",
+		            s_or_c (enroll),
 		            enroll->computer_sam, krb5_get_error_message (k5, code));
 		/* TODO: Parse out these values */
 		res = ADCLI_ERR_DIRECTORY;
@@ -1213,19 +1413,6 @@ retrieve_computer_account (adcli_enroll *enroll)
 	char *end;
 	int ret;
 
-	char *attrs[] =  {
-		"msDS-KeyVersionNumber",
-		"msDS-supportedEncryptionTypes",
-		"dNSHostName",
-		"servicePrincipalName",
-		"operatingSystem",
-		"operatingSystemVersion",
-		"operatingSystemServicePack",
-		"pwdLastSet",
-		"userAccountControl",
-		NULL,
-	};
-
 	assert (enroll->computer_dn != NULL);
 	assert (enroll->computer_attributes == NULL);
 
@@ -1233,12 +1420,14 @@ retrieve_computer_account (adcli_enroll *enroll)
 	assert (ldap != NULL);
 
 	ret = ldap_search_ext_s (ldap, enroll->computer_dn, LDAP_SCOPE_BASE,
-	                         "(objectClass=*)", attrs, 0, NULL, NULL, NULL, -1,
+	                         "(objectClass=*)", default_ad_ldap_attrs,
+	                         0, NULL, NULL, NULL, -1,
 	                         &enroll->computer_attributes);
 
 	if (ret != LDAP_SUCCESS) {
 		return _adcli_ldap_handle_failure (ldap, ADCLI_ERR_DIRECTORY,
-		                                   "Couldn't retrieve computer account info: %s",
+		                                   "Couldn't retrieve %s account info: %s",
+		                                   s_or_c (enroll),
 		                                   enroll->computer_dn);
 	}
 
@@ -1248,15 +1437,15 @@ retrieve_computer_account (adcli_enroll *enroll)
 		if (value != NULL) {
 			kvno = strtoul (value, &end, 10);
 			if (end == NULL || *end != '\0') {
-				_adcli_err ("Invalid kvno '%s' for computer account in directory: %s",
-				            value, enroll->computer_dn);
+				_adcli_err ("Invalid kvno '%s' for %s account in directory: %s",
+				            value, s_or_c (enroll), enroll->computer_dn);
 				res = ADCLI_ERR_DIRECTORY;
 
 			} else {
 				enroll->kvno = kvno;
 
-				_adcli_info ("Retrieved kvno '%s' for computer account in directory: %s",
-				             value, enroll->computer_dn);
+				_adcli_info ("Retrieved kvno '%s' for %s account in directory: %s",
+				             value, s_or_c (enroll), enroll->computer_dn);
 			}
 
 			free (value);
@@ -1265,8 +1454,8 @@ retrieve_computer_account (adcli_enroll *enroll)
 			/* Apparently old AD didn't have this attribute, use zero */
 			enroll->kvno = 0;
 
-			_adcli_info ("No kvno found for computer account in directory: %s",
-			             enroll->computer_dn);
+			_adcli_info ("No kvno found for %s account in directory: %s",
+			             s_or_c (enroll), enroll->computer_dn);
 		}
 	}
 
@@ -1307,12 +1496,14 @@ update_and_calculate_enctypes (adcli_enroll *enroll)
 
 	if (ret == LDAP_INSUFFICIENT_ACCESS) {
 		return _adcli_ldap_handle_failure (ldap, ADCLI_ERR_CREDENTIALS,
-		                                   "Insufficient permissions to set encryption types on computer account: %s",
+		                                   "Insufficient permissions to set encryption types on %s account: %s",
+		                                   s_or_c (enroll),
 		                                   enroll->computer_dn);
 
 	} else if (ret != LDAP_SUCCESS) {
 		return _adcli_ldap_handle_failure (ldap, ADCLI_ERR_DIRECTORY,
-		                                   "Couldn't set encryption types on computer account: %s",
+		                                   "Couldn't set encryption types on %s account: %s",
+		                                   s_or_c (enroll),
 		                                   enroll->computer_dn);
 	}
 
@@ -1335,13 +1526,14 @@ update_computer_attribute (adcli_enroll *enroll,
 	string = _adcli_ldap_mods_to_string (mods);
 	return_unexpected_if_fail (string != NULL);
 
-	_adcli_info ("Modifying computer account: %s", string);
+	_adcli_info ("Modifying %s account: %s", s_or_c (enroll), string);
 
 	ret = ldap_modify_ext_s (ldap, enroll->computer_dn, mods, NULL, NULL);
 
 	if (ret != LDAP_SUCCESS) {
-		_adcli_warn ("Couldn't set %s on computer account: %s: %s",
-		             string, enroll->computer_dn, ldap_err2string (ret));
+		_adcli_warn ("Couldn't set %s on %s account: %s: %s",
+		             string, s_or_c (enroll), enroll->computer_dn,
+		             ldap_err2string (ret));
 		res = ADCLI_ERR_DIRECTORY;
 	}
 
@@ -1365,8 +1557,8 @@ static char *get_user_account_control (adcli_enroll *enroll)
 
 		attr_val = strtoul (uac_str, &end, 10);
 		if (*end != '\0' || attr_val > UINT32_MAX) {
-			_adcli_warn ("Invalid userAccountControl '%s' for computer account in directory: %s, assuming 0",
-			            uac_str, enroll->computer_dn);
+			_adcli_warn ("Invalid userAccountControl '%s' for %s account in directory: %s, assuming 0",
+			            uac_str, s_or_c (enroll), enroll->computer_dn);
 		} else {
 			uac = attr_val;
 		}
@@ -1377,10 +1569,20 @@ static char *get_user_account_control (adcli_enroll *enroll)
 		uac = UAC_WORKSTATION_TRUST_ACCOUNT | UAC_DONT_EXPIRE_PASSWORD;
 	}
 
-	if (adcli_enroll_get_trusted_for_delegation (enroll)) {
-		uac |= UAC_TRUSTED_FOR_DELEGATION;
-	} else {
-		uac &= ~(UAC_TRUSTED_FOR_DELEGATION);
+	if (enroll->trusted_for_delegation_explicit) {
+		if (adcli_enroll_get_trusted_for_delegation (enroll)) {
+			uac |= UAC_TRUSTED_FOR_DELEGATION;
+		} else {
+			uac &= ~(UAC_TRUSTED_FOR_DELEGATION);
+		}
+	}
+
+	if (enroll->account_disable_explicit) {
+		if (adcli_enroll_get_account_disable (enroll)) {
+			uac |= UAC_ACCOUNTDISABLE;
+		} else {
+			uac &= ~(UAC_ACCOUNTDISABLE);
+		}
 	}
 
 	if (asprintf (&uac_str, "%d", uac) < 0) {
@@ -1395,23 +1597,37 @@ update_computer_account (adcli_enroll *enroll)
 {
 	int res = 0;
 	LDAP *ldap;
+	char *value = NULL;
+
+	/* No updates for service accounts */
+	if (enroll->is_service) {
+		return;
+	}
 
 	ldap = adcli_conn_get_ldap_connection (enroll->conn);
 	return_if_fail (ldap != NULL);
 
 	/* Only update attributes which are explicitly given on the command
-	 * line. Otherwise 'adcli update' must be always called with the same
-	 * set of options to make sure existing attributes are not deleted or
-	 * overwritten with different values. */
-	if (enroll->host_fqdn_explicit) {
+	 * line or not set in the existing AD object. Otherwise 'adcli update'
+	 * must be always called with the same set of options to make sure
+	 * existing attributes are not deleted or overwritten with different
+	 * values. */
+	if (enroll->computer_attributes != NULL) {
+		value = _adcli_ldap_parse_value (ldap,
+		                                 enroll->computer_attributes,
+		                                 "dNSHostName");
+	}
+	if (enroll->host_fqdn_explicit || value == NULL ) {
 		char *vals_dNSHostName[] = { enroll->host_fqdn, NULL };
 		LDAPMod dNSHostName = { LDAP_MOD_REPLACE, "dNSHostName", { vals_dNSHostName, } };
 		LDAPMod *mods[] = { &dNSHostName, NULL };
 
 		res |= update_computer_attribute (enroll, ldap, mods);
 	}
+	free (value);
 
-	if (res == ADCLI_SUCCESS && enroll->trusted_for_delegation_explicit) {
+	if (res == ADCLI_SUCCESS && (enroll->trusted_for_delegation_explicit ||
+	                             enroll->account_disable_explicit)) {
 		char *vals_userAccountControl[] = { NULL , NULL };
 		LDAPMod userAccountControl = { LDAP_MOD_REPLACE, "userAccountControl", { vals_userAccountControl, } };
 		LDAPMod *mods[] = { &userAccountControl, NULL };
@@ -1457,6 +1673,14 @@ update_computer_account (adcli_enroll *enroll)
 		res |= update_computer_attribute (enroll, ldap, mods);
 	}
 
+	if (res == ADCLI_SUCCESS && enroll->description != NULL) {
+		char *vals_description[] = { enroll->description, NULL };
+		LDAPMod description = { LDAP_MOD_REPLACE, "description", { vals_description, }, };
+		LDAPMod *mods[] = { &description, NULL, };
+
+		res |= update_computer_attribute (enroll, ldap, mods);
+	}
+
 	if (res != 0)
 		_adcli_info ("Updated existing computer account: %s", enroll->computer_dn);
 }
@@ -1469,6 +1693,11 @@ update_service_principals (adcli_enroll *enroll)
 	LDAP *ldap;
 	int ret;
 
+	/* No updates for service accounts */
+	if (enroll->is_service) {
+		return ADCLI_SUCCESS;
+	}
+
 	ldap = adcli_conn_get_ldap_connection (enroll->conn);
 	return_unexpected_if_fail (ldap != NULL);
 
@@ -1581,7 +1810,10 @@ load_keytab_entry (krb5_context k5,
 		    _adcli_str_has_suffix (name, "$") && !strchr (name, '/')) {
 			enroll->computer_name = name;
 			name[len - 1] = '\0';
-			_adcli_info ("Found computer name in keytab: %s", name);
+			_adcli_info ("Found %s name in keytab: %s",
+			             s_or_c (enroll), name);
+			adcli_conn_set_computer_name (enroll->conn,
+			                              enroll->computer_name);
 			name = NULL;
 
 		} else if (!enroll->host_fqdn && _adcli_str_has_prefix (name, "host/") && strchr (name, '.')) {
@@ -1970,17 +2202,27 @@ adcli_enroll_prepare (adcli_enroll *enroll,
 
 	adcli_clear_last_error ();
 
-	/* Basic discovery and figuring out enroll params */
-	res = ensure_host_fqdn (res, enroll);
-	res = ensure_computer_name (res, enroll);
-	res = ensure_computer_sam (res, enroll);
-	res = ensure_user_principal (res, enroll);
-	res = ensure_computer_password (res, enroll);
-	if (!(flags & ADCLI_ENROLL_NO_KEYTAB))
+	if (enroll->is_service) {
+		/* Ensure basic params for service accounts */
+		res = ensure_host_fqdn (res, enroll);
+		res = ensure_computer_name (res, enroll);
+		res = ensure_computer_sam (res, enroll);
+		res = ensure_computer_password (res, enroll);
 		res = ensure_host_keytab (res, enroll);
-	res = ensure_service_names (res, enroll);
-	res = ensure_service_principals (res, enroll);
-	res = ensure_keytab_principals (res, enroll);
+		res = ensure_keytab_principals (res, enroll);
+	} else {
+		/* Basic discovery and figuring out enroll params */
+		res = ensure_host_fqdn (res, enroll);
+		res = ensure_computer_name (res, enroll);
+		res = ensure_computer_sam (res, enroll);
+		res = ensure_user_principal (res, enroll);
+		res = ensure_computer_password (res, enroll);
+		if (!(flags & ADCLI_ENROLL_NO_KEYTAB))
+			res = ensure_host_keytab (res, enroll);
+		res = ensure_service_names (res, enroll);
+		res = ensure_service_principals (res, enroll);
+		res = ensure_keytab_principals (res, enroll);
+	}
 
 	return res;
 }
@@ -2125,6 +2367,62 @@ enroll_join_or_update_tasks (adcli_enroll *enroll,
 	return update_keytab_for_principals (enroll, flags);
 }
 
+static adcli_result
+adcli_enroll_add_description_for_service_account (adcli_enroll *enroll)
+{
+	const char *fqdn;
+	char *desc;
+
+	fqdn = adcli_conn_get_host_fqdn (enroll->conn);
+	return_unexpected_if_fail (fqdn != NULL);
+	if (asprintf (&desc, "Please do not edit, Service account for %s, "
+	                     "managed by adcli.", fqdn) < 0) {
+		return_unexpected_if_reached ();
+	}
+
+	adcli_enroll_set_description (enroll, desc);
+	free (desc);
+
+	return ADCLI_SUCCESS;
+}
+
+adcli_result
+adcli_enroll_add_keytab_for_service_account (adcli_enroll *enroll)
+{
+	adcli_result res;
+	krb5_context k5;
+	krb5_error_code code;
+	char def_keytab_name[MAX_KEYTAB_NAME_LEN];
+	char *lc_dom_name;
+	int ret;
+
+	if (adcli_enroll_get_keytab_name (enroll) == NULL) {
+		res = _adcli_krb5_init_context (&k5);
+		if (res != ADCLI_SUCCESS) {
+			return res;
+		}
+
+		code = krb5_kt_default_name (k5, def_keytab_name,
+		                             sizeof (def_keytab_name));
+		krb5_free_context (k5);
+		return_unexpected_if_fail (code == 0);
+
+		lc_dom_name = strdup (adcli_conn_get_domain_name (enroll->conn));
+		return_unexpected_if_fail (lc_dom_name != NULL);
+		_adcli_str_down (lc_dom_name);
+
+
+		ret = asprintf (&enroll->keytab_name, "%s.%s", def_keytab_name,
+		                                             lc_dom_name);
+		free (lc_dom_name);
+		return_unexpected_if_fail (ret > 0);
+	}
+
+	_adcli_info ("Using service account keytab: %s", enroll->keytab_name);
+
+	return ADCLI_SUCCESS;
+}
+
 adcli_result
 adcli_enroll_join (adcli_enroll *enroll,
                    adcli_enroll_flags flags)
@@ -2140,7 +2438,11 @@ adcli_enroll_join (adcli_enroll *enroll,
 	if (res != ADCLI_SUCCESS)
 		return res;
 
-	res = ensure_default_service_names (enroll);
+	if (enroll->is_service) {
+		res = adcli_enroll_add_description_for_service_account (enroll);
+	} else {
+		res = ensure_default_service_names (enroll);
+	}
 	if (res != ADCLI_SUCCESS)
 		return res;
 
@@ -2179,12 +2481,11 @@ adcli_enroll_load (adcli_enroll *enroll)
 }
 
 adcli_result
-adcli_enroll_update (adcli_enroll *enroll,
-		     adcli_enroll_flags flags)
+adcli_enroll_read_computer_account (adcli_enroll *enroll,
+		                    adcli_enroll_flags flags)
 {
 	adcli_result res = ADCLI_SUCCESS;
 	LDAP *ldap;
-	char *value;
 
 	return_unexpected_if_fail (enroll != NULL);
 
@@ -2204,17 +2505,29 @@ adcli_enroll_update (adcli_enroll *enroll,
 
 	/* Find the computer dn */
 	if (!enroll->computer_dn) {
-		res = locate_computer_account (enroll, ldap, NULL, NULL);
+		res = locate_computer_account (enroll, ldap, false, NULL, NULL);
 		if (res != ADCLI_SUCCESS)
 			return res;
 		if (!enroll->computer_dn) {
-			_adcli_err ("No computer account for %s exists", enroll->computer_sam);
+			_adcli_err ("No %s account for %s exists",
+			            s_or_c (enroll), enroll->computer_sam);
 			return ADCLI_ERR_CONFIG;
 		}
 	}
 
 	/* Get information about the computer account */
-	res = retrieve_computer_account (enroll);
+	return retrieve_computer_account (enroll);
+}
+
+adcli_result
+adcli_enroll_update (adcli_enroll *enroll,
+		     adcli_enroll_flags flags)
+{
+	adcli_result res = ADCLI_SUCCESS;
+	LDAP *ldap;
+	char *value;
+
+	res = adcli_enroll_read_computer_account (enroll, flags);
 	if (res != ADCLI_SUCCESS)
 		return res;
 
@@ -2239,9 +2552,43 @@ adcli_enroll_update (adcli_enroll *enroll,
 	}
 	free (value);
 
+	/* We only support password changes for service accounts */
+	if (enroll->is_service && (flags & ADCLI_ENROLL_PASSWORD_VALID)) {
+		return ADCLI_SUCCESS;
+	}
+
 	return enroll_join_or_update_tasks (enroll, flags);
 }
 
+adcli_result
+adcli_enroll_show_computer_attribute (adcli_enroll *enroll)
+{
+	LDAP *ldap;
+	size_t c;
+	char **vals;
+	size_t v;
+
+	ldap = adcli_conn_get_ldap_connection (enroll->conn);
+	assert (ldap != NULL);
+
+	for (c = 0; default_ad_ldap_attrs[c] != NULL; c++) {
+		vals = _adcli_ldap_parse_values (ldap,
+		                                 enroll->computer_attributes,
+		                                 default_ad_ldap_attrs[c]);
+		printf ("%s:\n", default_ad_ldap_attrs[c]);
+		if (vals == NULL) {
+			printf (" - not set -\n");
+		} else {
+			for (v = 0; vals[v] != NULL; v++) {
+				printf (" %s\n", vals[v]);
+			}
+		}
+		_adcli_strv_free (vals);
+	}
+
+	return ADCLI_SUCCESS;
+}
+
 adcli_result
 adcli_enroll_delete (adcli_enroll *enroll,
                      adcli_enroll_flags delete_flags)
@@ -2271,11 +2618,12 @@ adcli_enroll_delete (adcli_enroll *enroll,
 
 	/* Find the computer dn */
 	if (!enroll->computer_dn) {
-		res = locate_computer_account (enroll, ldap, NULL, NULL);
+		res = locate_computer_account (enroll, ldap, false, NULL, NULL);
 		if (res != ADCLI_SUCCESS)
 			return res;
 		if (!enroll->computer_dn) {
-			_adcli_err ("No computer account for %s exists",
+			_adcli_err ("No %s account for %s exists",
+			            s_or_c (enroll),
 			            enroll->computer_sam);
 			return ADCLI_ERR_CONFIG;
 		}
@@ -2314,11 +2662,12 @@ adcli_enroll_password (adcli_enroll *enroll,
 
 	/* Find the computer dn */
 	if (!enroll->computer_dn) {
-		res = locate_computer_account (enroll, ldap, NULL, NULL);
+		res = locate_computer_account (enroll, ldap, false, NULL, NULL);
 		if (res != ADCLI_SUCCESS)
 			return res;
 		if (!enroll->computer_dn) {
-			_adcli_err ("No computer account for %s exists",
+			_adcli_err ("No %s account for %s exists",
+			            s_or_c (enroll),
 			            enroll->computer_sam);
 			return ADCLI_ERR_CONFIG;
 		}
@@ -2857,6 +3206,55 @@ adcli_enroll_set_trusted_for_delegation (adcli_enroll *enroll,
 	enroll->trusted_for_delegation_explicit = 1;
 }
 
+bool
+adcli_enroll_get_account_disable (adcli_enroll *enroll)
+{
+	return_val_if_fail (enroll != NULL, false);
+
+	return enroll->account_disable;
+}
+
+void
+adcli_enroll_set_account_disable (adcli_enroll *enroll,
+                                  bool value)
+{
+	return_if_fail (enroll != NULL);
+
+	enroll->account_disable = value;
+	enroll->account_disable_explicit = 1;
+}
+
+void
+adcli_enroll_set_description (adcli_enroll *enroll, const char *value)
+{
+	return_if_fail (enroll != NULL);
+	if (value != NULL && value[0] != '\0') {
+		_adcli_str_set (&enroll->description, value);
+	}
+}
+
+const char *
+adcli_enroll_get_desciption (adcli_enroll *enroll)
+{
+	return_val_if_fail (enroll != NULL, NULL);
+	return enroll->description;
+}
+
+void
+adcli_enroll_set_is_service (adcli_enroll *enroll, bool value)
+{
+	return_if_fail (enroll != NULL);
+
+	enroll->is_service = value;
+	enroll->is_service_explicit = true;
+}
+
+bool
+adcli_enroll_get_is_service (adcli_enroll *enroll)
+{
+	return enroll->is_service;
+}
+
 const char **
 adcli_enroll_get_service_principals_to_add (adcli_enroll *enroll)
 {
